@@ -61,6 +61,9 @@ macro_rules! question_mark {
     }};
 }
 
+/// Syscall initialization function
+pub type SyscallInit<'a, C, E> = fn(C) -> Box<(dyn SyscallObject<E> + 'a)>;
+
 /// Syscall function without context
 pub type SyscallFunction<E, O> =
     fn(O, u64, u64, u64, u64, u64, &MemoryMapping, &mut ProgramResult<E>);
@@ -84,6 +87,8 @@ pub trait SyscallObject<E: UserDefinedError> {
 /// Syscall function and binding slot for a context object
 #[derive(Debug, PartialEq)]
 pub struct Syscall {
+    /// Syscall init
+    pub init: u64,
     /// Call the syscall function
     pub function: u64,
     /// Slot of context object
@@ -123,11 +128,13 @@ pub struct SyscallRegistry {
 
 impl SyscallRegistry {
     /// Register a syscall function by its symbol hash
-    pub fn register_syscall_by_hash<E: UserDefinedError, O: SyscallObject<E>>(
+    pub fn register_syscall_by_hash<'a, C, E: UserDefinedError, O: SyscallObject<E>>(
         &mut self,
         hash: u32,
+        init: SyscallInit<'a, C, E>,
         function: SyscallFunction<E, &mut O>,
     ) -> Result<(), EbpfError<E>> {
+        let init = init as *const u8 as u64;
         let function = function as *const u8 as u64;
         let context_object_slot = self.entries.len();
         if self
@@ -135,6 +142,7 @@ impl SyscallRegistry {
             .insert(
                 hash,
                 Syscall {
+                    init,
                     function,
                     context_object_slot,
                 },
@@ -152,12 +160,13 @@ impl SyscallRegistry {
     }
 
     /// Register a syscall function by its symbol name
-    pub fn register_syscall_by_name<E: UserDefinedError, O: SyscallObject<E>>(
+    pub fn register_syscall_by_name<'a, C, E: UserDefinedError, O: SyscallObject<E>>(
         &mut self,
         name: &[u8],
+        init: SyscallInit<'a, C, E>,
         function: SyscallFunction<E, &mut O>,
     ) -> Result<(), EbpfError<E>> {
-        self.register_syscall_by_hash(ebpf::hash_symbol_name(name), function)
+        self.register_syscall_by_hash::<C, E, O>(ebpf::hash_symbol_name(name), init, function)
     }
 
     /// Get a symbol's function pointer and context object slot
@@ -222,6 +231,11 @@ pub struct Config {
     pub enable_sdiv: bool,
     /// Avoid copying read only sections when possible
     pub optimize_rodata: bool,
+    /// Support syscalls via pseudo calls (insn.src = 0)
+    pub static_syscalls: bool,
+    /// Allow sh_addr != sh_offset in elf sections. Used in SBFv2 to align
+    /// section vaddrs to MM_PROGRAM_START.
+    pub enable_elf_vaddr: bool,
 }
 
 impl Config {
@@ -249,9 +263,11 @@ impl Default for Config {
             disable_deprecated_load_instructions: true,
             syscall_bpf_function_hash_collision: true,
             reject_callx_r10: true,
-            dynamic_stack_frames: false,
+            dynamic_stack_frames: true,
             enable_sdiv: true,
             optimize_rodata: true,
+            static_syscalls: true,
+            enable_elf_vaddr: true,
         }
     }
 }
@@ -568,7 +584,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
         &self.tracer
     }
 
-    /// Bind a context object instance to a previously registered syscall
+    /// Initializes and binds the context object instances for all previously registered syscalls
     ///
     /// # Examples
     ///
@@ -594,7 +610,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// // On running the program this syscall will print the content of registers r3, r4 and r5 to
     /// // standard output.
     /// let mut syscall_registry = SyscallRegistry::default();
-    /// syscall_registry.register_syscall_by_hash(6, BpfTracePrintf::call).unwrap();
+    /// syscall_registry.register_syscall_by_hash(6, BpfTracePrintf::init::<u64, UserError>, BpfTracePrintf::call).unwrap();
     /// // Instantiate an Executable and VM
     /// let config = Config::default();
     /// let mut bpf_functions = std::collections::BTreeMap::new();
@@ -602,37 +618,37 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// let mut executable = Executable::<UserError, TestInstructionMeter>::from_text_bytes(prog, None, config, syscall_registry, bpf_functions).unwrap();
     /// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(&executable, &mut [], Vec::new()).unwrap();
     /// // Bind a context object instance to the previously registered syscall
-    /// vm.bind_syscall_context_object(Box::new(BpfTracePrintf {}), None);
+    /// vm.bind_syscall_context_objects(0);
     /// ```
-    pub fn bind_syscall_context_object(
+    pub fn bind_syscall_context_objects<C: Clone>(
         &mut self,
-        syscall_context_object: Box<dyn SyscallObject<E> + 'a>,
-        hash: Option<u32>,
+        syscall_context: C,
     ) -> Result<(), EbpfError<E>> {
-        let fat_ptr: DynTraitFatPointer = unsafe { std::mem::transmute(&*syscall_context_object) };
         let syscall_registry = self.executable.get_syscall_registry();
-        let slot = match hash {
-            Some(hash) => {
-                syscall_registry
-                    .lookup_syscall(hash)
-                    .ok_or(EbpfError::SyscallNotRegistered(hash as usize))?
-                    .context_object_slot
-            }
-            None => syscall_registry
+
+        for syscall in syscall_registry.entries.values() {
+            let syscall_object_init_fn: SyscallInit<C, E> =
+                unsafe { std::mem::transmute(syscall.init) };
+            let syscall_context_object: Box<dyn SyscallObject<E> + 'a> =
+                syscall_object_init_fn(syscall_context.clone());
+            let fat_ptr: DynTraitFatPointer =
+                unsafe { std::mem::transmute(&*syscall_context_object) };
+            let slot = syscall_registry
                 .lookup_context_object_slot(fat_ptr.vtable.methods[0] as u64)
                 .ok_or(EbpfError::SyscallNotRegistered(
                     fat_ptr.vtable.methods[0] as usize,
-                ))?,
-        };
-        if !self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + slot].is_null() {
-            Err(EbpfError::SyscallAlreadyBound(slot))
-        } else {
+                ))?;
+
+            debug_assert!(
+                self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + slot].is_null()
+            );
             self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + slot] = fat_ptr.data;
             // Keep the dyn trait objects so that they can be dropped properly later
             self.syscall_context_object_pool
                 .push(syscall_context_object);
-            Ok(())
         }
+
+        Ok(())
     }
 
     /// Lookup a syscall context object by its function pointer. Used for testing and validation.
@@ -702,9 +718,11 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
         }
 
         // Check config outside of the instruction loop
-        let instruction_meter_enabled = self.executable.get_config().enable_instruction_meter;
-        let instruction_tracing_enabled = self.executable.get_config().enable_instruction_tracing;
-        let dynamic_stack_frames = self.executable.get_config().dynamic_stack_frames;
+        let config = self.executable.get_config();
+        let instruction_meter_enabled = config.enable_instruction_meter;
+        let instruction_tracing_enabled = config.enable_instruction_tracing;
+        let dynamic_stack_frames = config.dynamic_stack_frames;
+        let static_syscalls = config.static_syscalls;
 
         // Loop on instructions
         let entry = self.executable.get_entrypoint_instruction_offset()?;
@@ -1022,36 +1040,57 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                 // Do not delegate the check to the verifier, since registered functions can be
                 // changed after the program has been verified.
                 ebpf::CALL_IMM => {
-                    if let Some(syscall) = self.executable.get_syscall_registry().lookup_syscall(insn.imm as u32) {
-                        if instruction_meter_enabled {
-                            let _ = instruction_meter.consume(self.last_insn_count);
-                        }
-                        total_insn_count += self.last_insn_count;
-                        self.last_insn_count = 0;
-                        let mut result: ProgramResult<E> = Ok(0);
-                        (unsafe { std::mem::transmute::<u64, SyscallFunction::<E, *mut u8>>(syscall.function) })(
-                            self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + syscall.context_object_slot],
-                            reg[1],
-                            reg[2],
-                            reg[3],
-                            reg[4],
-                            reg[5],
-                            &self.memory_mapping,
-                            &mut result,
-                        );
-                        reg[0] = result?;
-                        if instruction_meter_enabled {
-                            remaining_insn_count = instruction_meter.get_remaining();
-                        }
-                    } else if let Some(target_pc) = self.executable.lookup_bpf_function(insn.imm as u32) {
-                        // make BPF to BPF call
-                        reg[ebpf::FRAME_PTR_REG] =
-                            self.stack.push(&reg[ebpf::FIRST_SCRATCH_REG..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS], next_pc)?;
-                        next_pc = self.check_pc(pc, target_pc)?;
-                    } else if self.executable.get_config().disable_unresolved_symbols_at_runtime {
-                        return Err(EbpfError::UnsupportedInstruction(pc + ebpf::ELF_INSN_DUMP_OFFSET));
+                    let mut resolved = false;
+                    let (syscalls, calls) = if static_syscalls {
+                        (insn.src == 0, insn.src != 0)
                     } else {
-                        self.executable.report_unresolved_symbol(pc)?;
+                        (true, true)
+                    };
+
+                    if syscalls {
+                        if let Some(syscall) = self.executable.get_syscall_registry().lookup_syscall(insn.imm as u32) {
+                            resolved = true;
+
+                            if instruction_meter_enabled {
+                                let _ = instruction_meter.consume(self.last_insn_count);
+                            }
+                            total_insn_count += self.last_insn_count;
+                            self.last_insn_count = 0;
+                            let mut result: ProgramResult<E> = Ok(0);
+                            (unsafe { std::mem::transmute::<u64, SyscallFunction::<E, *mut u8>>(syscall.function) })(
+                                self.syscall_context_objects[SYSCALL_CONTEXT_OBJECTS_OFFSET + syscall.context_object_slot],
+                                reg[1],
+                                reg[2],
+                                reg[3],
+                                reg[4],
+                                reg[5],
+                                &self.memory_mapping,
+                                &mut result,
+                            );
+                            reg[0] = result?;
+                            if instruction_meter_enabled {
+                                remaining_insn_count = instruction_meter.get_remaining();
+                            }
+                        }
+                    }
+
+                    if calls {
+                        if let Some(target_pc) = self.executable.lookup_bpf_function(insn.imm as u32) {
+                            resolved = true;
+
+                            // make BPF to BPF call
+                            reg[ebpf::FRAME_PTR_REG] =
+                                self.stack.push(&reg[ebpf::FIRST_SCRATCH_REG..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS], next_pc)?;
+                            next_pc = self.check_pc(pc, target_pc)?;
+                        }
+                    }
+
+                    if !resolved {
+                        if config.disable_unresolved_symbols_at_runtime {
+                            return Err(EbpfError::UnsupportedInstruction(pc + ebpf::ELF_INSN_DUMP_OFFSET));
+                        } else {
+                            self.executable.report_unresolved_symbol(pc)?;
+                        }
                     }
                 }
 
